/*
 * TRANANAL.C
 *
 * Extract information from listing generated by VS/COBOL II MAP option and
 * Expert System mapping database extract to create a translation request
 * list, written to standard output.
 * 
 * It is expected that a COBOL shell program is used that includes
 * an FD for which a DMAP is desired.
 * 
 * Usage:
 * 
 *    trananal <dmap> <dmi> > <tranlist>
 * 
 * where
 * 
 * 	<dmap>      is the VS/COBOL II Data Map listing
 * 	<dmi>       is an extract from the Expert System mapping database that
 *                indicates DM attribute names and types, corresponding COBOL
 *                record fields, and special translation processing, if such
 *                is required
 *    <tranlist>  the translation request list; lists DM input fields, types,
 *                mapping, and attributes (special translation processing)
*/

#include "env.h"  /* must appear before first C statement */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hash.h"
#include "dmio.h"
#include "tran.h"

#define _MAX_FNAME 256

typedef struct
{
	char *typeName;
	cobolType ct;
} cobolTypeData;

/* initialized to 0 */
int convFuncCt[(int) unknownCobolType + 1][(int) unknownDmType + 1]
	[(int) unknownAttribute + 1];

char *progname, line[MAX_LINE];

void usage(void);

bool getDisp(char **ppc, unsigned long *pulHexNum);
bool getHexNumber(char **ppc, unsigned long *pulHexNum);
bool skipLeadingPeriods(char **ppc);

cobolType getType(char *strType);
char *getTypeName(cobolType t);
ulong getLength(char *strDefinition);
dmType getDmType(char *strDmType);
void printStEntry(symtab s, stEntry *pstEntry);


/* number of incidences of each type, index by cobolType */
static int cobolTypeUsage[7] = { 0 };

static cobolTypeData cobolTypeTable[] =
{
	"DISPLAY",	alphaCobolType,
	"COMP-2",	doubleCobolType,
	"COMP-1",	floatCobolType,
	"COMP",		integerCobolType,
	"DISP-NUM", numericTextCobolType,
	"NUM-EDIT", numericTextEditedCobolType,
	"COMP-3",	packedCobolType,
	"UNKNOWN",	unknownCobolType,
	"",			unknownCobolType
};

typedef struct
{
	char *typeName;
	dmType dmt;
} dmTypeData;

dmTypeData dmTypeTable[] =
{
	"F",	floatDmType,
	"I",	integerDmType,
	"L",	logicalDmType,
	"S",	stringDmType,
	"U",	unknownDmType,
	"",	unknownDmType
};

typedef enum
{
	alphaToStrCFunc, copyCFunc, packToFloatCFunc, packToIntCFunc,
	packToStrCFunc, zoneToFloatCFunc, zoneToIntCFunc, unsupportedCFunc
} CFunc;

/*
	convFuncs maps CobolTypes and DmTypes to convFuncs.
	unknown types are always unsupported.
Note: numericText may need decimal place when converting to String or Logical.
	Force x(6) ?
 */
CFunc convFuncs[(int) unknownCobolType + 1][(int) unknownDmType + 1] =
{
 /* 	F					I					L					S					U			*/
	{ unsupportedCFunc,	copyCFunc,			copyCFunc,			alphaToStrCFunc,
 		unsupportedCFunc },
	{ unsupportedCFunc,	unsupportedCFunc,	unsupportedCFunc,	unsupportedCFunc,
 		unsupportedCFunc },
	{ unsupportedCFunc,	unsupportedCFunc,	unsupportedCFunc,	unsupportedCFunc,
		unsupportedCFunc },
	{ unsupportedCFunc,	unsupportedCFunc,	unsupportedCFunc,	unsupportedCFunc,
		unsupportedCFunc },
	{ zoneToFloatCFunc,	zoneToIntCFunc,	copyCFunc,			alphaToStrCFunc,
		unsupportedCFunc },
	{ unsupportedCFunc,	unsupportedCFunc,	unsupportedCFunc,	unsupportedCFunc,
		unsupportedCFunc },
	{ packToFloatCFunc,	packToIntCFunc,	unsupportedCFunc,	packToStrCFunc, 
		unsupportedCFunc },
	{ unsupportedCFunc,	unsupportedCFunc,	unsupportedCFunc,	unsupportedCFunc,
		unsupportedCFunc }
};

attributeType alphaToStrCFuncAttributes[] =
{
	keyedStringAttribute,
	tableAttribute,
	keyedStringIfAttribute,
	keyedStringTableAttribute,
	keyedStringIf2Attribute,
	keyedStringIfIntAttribute,
	unknownAttribute
};

attributeType copyCFuncAttributes[] =
{
	dateAttribute,
	keyAttribute,
	tableAttribute,
	validateAttribute,
	validateIndirectOffAttribute,
	unknownAttribute
};

attributeType packToFloatCFuncAttributes[] =
{
	decimalPlacesAttribute,
	unknownAttribute
};

attributeType packToIntCFuncAttributes[] =
{
	byThousandAttribute,
	by12OffAttribute,
	toMonthsAttribute,
	keyedStringIfIntAttribute,
	unknownAttribute
};

attributeType packToStrCFuncAttributes[] =
{
	keyedStringAttribute,
	tableAttribute,
	keyedStringIfAttribute,
	keyedStringTableAttribute,
	keyedStringIf2Attribute,
	keyedStringIfIntAttribute,
	unknownAttribute
};

attributeType zoneToFloatCFuncAttributes[] =
{
	decimalPlacesAttribute,
	unknownAttribute
};

attributeType zoneToIntCFuncAttributes[] =
{
	byThousandAttribute,
	by12OffAttribute,
	toMonthsAttribute,
	keyedStringIfIntAttribute,
	unknownAttribute
};

attributeType unsupportedCFuncAttributes[] =
{
	unknownAttribute
};

typedef struct
{
	char *CFuncName;
	CFunc cf;
	attributeType *pat;
} CFuncDesc;

CFuncDesc cfdTable[] =
{
	"alphaToStr",	alphaToStrCFunc,	alphaToStrCFuncAttributes,
	"copy",			copyCFunc,			copyCFuncAttributes,
	"packToFloat",	packToFloatCFunc,	packToFloatCFuncAttributes,
	"packToInt",	packToIntCFunc,	packToIntCFuncAttributes,
	"packToStr",	packToStrCFunc,	packToStrCFuncAttributes,
	"zoneToFloat",	zoneToFloatCFunc,	zoneToFloatCFuncAttributes,
	"zoneToInt",	zoneToIntCFunc,	zoneToIntCFuncAttributes,
	"unsupported",	unsupportedCFunc,	unsupportedCFuncAttributes,
	"",				unsupportedCFunc,	unsupportedCFuncAttributes
};

int main(int argc, char *argv[]);

int main(int argc, char *argv[])
{
	symtab st;
	stEntry *pstEntry;
	FILE *fpDmap, *fpDminput;
	char strCobolName[MAX_COBOL_NAME + 1];
	char *pStart = line, *pEnd;
	bool parseSuccessful;
	int iLineCt, i, j, k, iRecLen = 0, iMaxIsamRecLen = 0, iMaxFldLen = 0;
	QUEUE q;
	char *pstr;
	int iLen;
	key theKey;
	/* strOutput must be large enough to hold an entire tran line */
	char strOutput[256];

   progname = argv[0];

   /* parse command line options */
   while ( --argc > 0  &&  (*++argv)[0] == '-' )
      switch((*argv)[1])
		{
      case '?':
		default:
			usage();
      }

	if(2 != argc)
		usage();

	if(NULL == (fpDmap = fopen(argv[0], "r")))
	{
		fprintf(stderr,"%s: cannot open '%s'\n", progname, argv[0]);
		return(1);
	}

	if(NULL == (fpDminput = fopen(argv[1], "r")))
	{
		fprintf(stderr,"%s: cannot open '%s'\n", progname, argv[1]);
		return(1);
	}

	st = dmstConstruct(MAXSYMS);
	while(fgets(line, sizeof(line), fpDmap))
	{
		unsigned long ulLine, ulLevel, ulSegDisp, ulDisp;
		char sBase[9], sDs[3], sDef[16], sUsage[16];
		cobolType ct;

		pStart = line;
		if(parseSuccessful)
			parseSuccessful = getCobolName(&pStart, strCobolName);
		if(parseSuccessful)
			parseSuccessful = getDecimalNumber(&pStart, &ulLevel);
		if(parseSuccessful)
			parseSuccessful = getDecimalNumber(&pStart, &ulLine);
		if(parseSuccessful)
			parseSuccessful = getDecimalNumber(&pStart, &ulDisp);
		if(parseSuccessful)
			parseSuccessful = getDecimalNumber(&pStart, &iLen);
		if(parseSuccessful)
			parseSuccessful = getString(&pStart, sUsage, sizeof(sUsage));
		if(parseSuccessful)
		{
			if(NULL != findsym(st, strCobolName))
				fprintf(stderr, "Redefinition of COBOL field %s\n", strCobolName);
			else
			{
				(void) install( st, strCobolName, ct = getType(sUsage), ulDisp,
					             (ulong) iLen);
				cobolTypeUsage[(int) ct]++;
				iRecLen = (int) ulDisp + iLen;
			}
		}
   }

#if 0
	dmstIterate(st, printStEntry);
#endif
	for(i = 0; i < 7; i++)
	{
		fprintf(stderr, "usage of type %s: %d\n",
			cobolTypeTable[i].typeName, cobolTypeUsage[i]);
	}

	/* !!! output VSAM file definition instructions */

	q = qcreate();
	for(iLineCt = 1; NULL != fgets(line, sizeof(line), fpDminput); iLineCt++)
	{
		char strDmType[2], strDmName[MAX_DM_NAME + 1];
		dmType dmt;
		attribute attr;
		int iAttrCt;
		CFunc cf;

		pStart = line;
		parseSuccessful = getString(&pStart, strDmType, sizeof(strDmType));
		if(parseSuccessful)
		{
			dmt = getDmType(strDmType);
			if(unknownDmType == dmt)
				parseSuccessful = false;
		}
		if(parseSuccessful)
			parseSuccessful = getDMName(&pStart, strDmName);
		if(parseSuccessful)
			parseSuccessful = getCobolName(&pStart, strCobolName);

		/* read optional conversion attributes */
		if(parseSuccessful)
		{
			iAttrCt = 0;
			/* attribute is optional */
			if(getAttribute(&pStart, &attr))
			{
				/* validate attribute information using .dmi info */
				switch(attr.type)
				{
				case keyAttribute:
					if(0 != strcmp(theKey.name, ""))
					{
						fprintf(stderr, "Duplicate key declaration %s and %s\n",
							theKey.name, strCobolName);
					}
					break;

				case tableAttribute:
					strlwr(attr.u.tb.tblName);
					if(NULL == getTable(attr.u.tb.tblName))
					{
						parseSuccessful = false;
						fprintf(stderr,
							"(%s) tableCon() references invalid table"
							" in line %d:\n\t %s",
							argv[1],	iLineCt, line);
					}
					break;

				case validateAttribute:
					strlwr(attr.u.val.tblName);
					if(NULL == getTable(attr.u.val.tblName))
					{
						parseSuccessful = false;
						fprintf(stderr,
							"(%s) validate() references invalid table"
							" in line %d:\n\t %s",
							argv[1],	iLineCt, line);
					}
					break;
				}
				iAttrCt++;
			}
		}

		/* check to see that line is completely parsed */
		if(parseSuccessful)
		{
			pEnd = pStart;
			while( *pEnd && isspace(*pEnd) )
				pEnd++;
			if(*pEnd)
				fprintf(stderr, "(%s) unrecognized attribute(s) in line %d: %s\n",
					argv[1], iLineCt, pStart);
		}

		if(!parseSuccessful || (NULL == (pstEntry = findsym(st, strCobolName))))
		{
			parseSuccessful = false;
			fprintf(stderr, "can't find COBOL field %s\n", strCobolName);
		}
		/* validate attribute information using .dmi & .dmp info
			convert field references to offset, length
		 */
		if(parseSuccessful && 0 < iAttrCt)
		{
			stEntry *pstField;

			switch(attr.type)
			{
			case by12Attribute:
				if(NULL == (pstField = findsym(st, attr.u.by12.field)))
				{
					fprintf(stderr,
						"(%s) by12() references invalid field in line %d:\n\t %s",
						argv[1],	iLineCt, line);
					parseSuccessful = false;
				}
				else
				{
					attr.type = by12OffAttribute;
					attr.u.by12Off.ulOff = pstField->offset;
					attr.u.by12Off.ulLen = pstField->length;
				}
				break;

			case dateAttribute:
				if(pstEntry->length != 6)
				{
					parseSuccessful = false;
					fprintf(stderr,
						"(%s) date field length != 6 in line %d:\n\t %s",
						argv[1],	iLineCt, line);
				}
				break;

			case keyAttribute:
				strcpy(theKey.name, strCobolName);
				theKey.ulLen = pstEntry->length;
				theKey.ulOffset = pstEntry->offset;
				break;

			case keyedStringAttribute:
			case indirectKeyedStringAttribute:
				if((int) pstEntry->length > iMaxIsamRecLen)
					iMaxIsamRecLen = (int) pstEntry->length;
				break;

			case validateIndirectAttribute:
				if(NULL == (pstField = findsym(st, attr.u.valind.field)))
				{
					fprintf(stderr,
						"(%s) validateIndirect() references invalid field"
						" in line %d:\n\t %s",
						argv[1],	iLineCt, line);
					parseSuccessful = false;
				}
				else
				{
					attr.type = validateIndirectOffAttribute;
					/* tblName is located at the same location in both valind and
						valindOff */
					/* this copy may overlap */
					strcpy(attr.u.valindOff.strElse, attr.u.valind.strElse);
					attr.u.valindOff.ulOffset = pstField->offset;
					attr.u.valindOff.ulLen = pstField->length;
				}
				break;

			default:
				break;
			}
			pstEntry->attr = attr;
		}
		if(parseSuccessful)
		{
			pstEntry->iAttrCt = iAttrCt;

			if(iMaxFldLen < (int) pstEntry->length)
				iMaxFldLen = (int) pstEntry->length;

			sprintf(strOutput, "%s %s %04ld %04ld", strDmName,
				hash_name(st, pstEntry), pstEntry->offset, pstEntry->length);
			/* check for compatibility of DM and COBOL datatypes,
				generate conversion code

				For now, write attribute name.
			 */
			cf = convFuncs[(int) pstEntry->type][(int) dmt];
			if(unsupportedCFunc == cf)
			{
				fprintf(stderr,
					"(%s) unsupported CFunc (cobolType %s, DMType %s)"
					" in line %d:\n\t %s",
					argv[1], cobolTypeTable[(int) pstEntry->type].typeName,
					dmTypeTable[(int) dmt].typeName,	iLineCt, line);
				parseSuccessful = false;
			}
			else if(0 < iAttrCt)
			{
				attributeType *pat;
				bool bFound = false;

				for(pat = cfdTable[(int) cf].pat; unknownAttribute != *pat; pat++)
				{
					if(attr.type == *pat)
					{
						bFound = true;
						break;
					}
				}
				if(!bFound)
				{
					parseSuccessful = false;
					fprintf(stderr,
						"(%s) unsupported attribute (CFunc %s, attribute %s)"
						" in line %d:\n\t %s",
						argv[1], cfdTable[(int) cf].CFuncName,
						attributeDescLookup(attr.type)->strArgName,
						iLineCt, line);
				}
				if(parseSuccessful)
					convFuncCt[(int) pstEntry->type][(int) dmt][(int) attr.type]++;
			}
		}

		if(parseSuccessful)
		{
			strcat(strOutput, " ");
			strcat(strOutput, cfdTable[(int) cf].CFuncName);

			for(i = 0; i < pstEntry->iAttrCt; i++)
			{
				strcat(strOutput, " ");
				attributeStoreOn(&pstEntry->attr, strOutput);
			}
			pstr = qadd(q, strlen(strOutput) + 1);
			strcpy(pstr, strOutput);
#if !defined(NDEBUG)
			if(numericTextCobolType == pstEntry->type && logicalDmType == dmt)
				fprintf(stderr, "DISP-NUM->L field %s attr %s\n", hash_name(st, pstEntry),
					attributeDescLookup(attr.type)->strArgName);
			if(numericTextCobolType == pstEntry->type && stringDmType == dmt)
				fprintf(stderr, "DISP-NUM->S field %s attr %s\n", hash_name(st, pstEntry),
					attributeDescLookup(attr.type)->strArgName);
#endif
		}
	}

	printf("%d %d %s %d %d %d\n", iRecLen, iMaxIsamRecLen, theKey.name,
		(int) theKey.ulOffset, (int) theKey.ulLen, iMaxFldLen + 1);
	FOR_EACH_Q_ELE(q, pstr, char *)
		printf("%s\n", pstr);
	qdestroy(&q);

	fprintf(stderr, "Conversion usage:\n");
	for(i = 0; i < (int) unknownCobolType; i++)
		for(j = 0; j < (int) unknownDmType; j++)
			if(unsupportedCFunc != convFuncs[i][j])
				for(k = 0; k < (int) unknownAttribute; k++)
				{
					attributeDesc *pad;

					if(0 < convFuncCt[i][j][k])
					{
						pad = attributeDescLookup((attributeType) k);
						fprintf(stderr, "%s->%s (%s) w/att %s \t%3d\n",
							cobolTypeTable[i].typeName,
							dmTypeTable[j].typeName,
							cfdTable[(int) convFuncs[i][j]].CFuncName,
							pad->strArgName,
							convFuncCt[i][j][k]);
					}
				}

	dmstDestruct(&st);
	fclose(fpDmap);
	fclose(fpDminput);
   return(0);
}

void printStEntry(symtab s, stEntry *pstEntry)
{
	printf("Nm %s  Type %s  Length %04ld  Offset %04ld\n",
		hash_name(s, pstEntry), getTypeName(pstEntry->type), pstEntry->length,
		pstEntry->offset);
}

dmType getDmType(char *strDmType)
{
	if(0 == strcmp(strDmType, "F"))
		return(floatDmType);
	else if(0 == strcmp(strDmType, "I"))
		return(integerDmType);
	else if(0 == strcmp(strDmType, "L"))
		return(logicalDmType);
	else if(0 == strcmp(strDmType, "S"))
		return(stringDmType);
	else
		return(unknownDmType);
}


void usage()
{
	fprintf(stderr, "usage: %s <dmap_filename> <dminput_filename> \n",
		progname);
	exit(1);
}


cobolType getType(char *strType)
{
	int i;
	bool found;

	for(i = 0, found = false;
		!found && 0 != strcmp("", cobolTypeTable[i].typeName); i++)
	{
		if(0 == strcmp(strType, cobolTypeTable[i].typeName))
			found = true;
	}
	if(found)
		return(cobolTypeTable[--i].ct);
	else
		return(unknownCobolType);
}


char *getTypeName(cobolType t)
{
	return(cobolTypeTable[(int) t].typeName);
}


/* getLength of a valid IBM data division map definition */
ulong getLength(char *strDefinition)
{
	char *pEnd = strDefinition;
	ulong ulLen;

	(void) getDecimalNumber(&pEnd, &ulLen);
	return(ulLen);
}


bool getHexNumber(char **ppc, unsigned long *pulHexNum)
{
	bool rc = false;
	char *p = *ppc;
	char stHexNumber[MAX_HEX_DIGITS + 1];
	int idx = 0;

	if(NULL == p)
		return(false);

	/* skip white space */
	while(*p && isspace(*p))
		p++;
	while(*p && isxdigit(*p))
	{
		if(idx < MAX_HEX_DIGITS)
		{
			rc = true;
			stHexNumber[idx++] = *p++;
		}
		else
			rc = false;
	}
	if(rc)
	{
		stHexNumber[idx] = EOS;
		sscanf(stHexNumber, "%lx", pulHexNum);
		*ppc = p;
	}

	return(rc);
}

bool getDisp(char **ppc, unsigned long *pulHexNum)
{
	bool rc = false;
	char *p = *ppc;
	char stHexNumber[MAX_HEX_DIGITS + 1];
	int idx = 0;
	bool bDone;

	if(NULL == p)
		return(false);

	/* skip white space */
	while(*p && isspace(*p))
		p++;


	for(bDone = false; *p && !bDone; p++)
	{
		if(isxdigit(*p))
		{
			if(idx < MAX_HEX_DIGITS)
			{
				rc = true;
				stHexNumber[idx++] = *p;
			}
			else
				rc = false;
		}
		else if(isspace(*p))
		{
			/* a single space between portions of the displacement is allowed */
			if(*(p+1) && isxdigit(*(p+1)))
				;
			else
				bDone = true;
		}
		else
			bDone = true;
	}
	if(rc)
	{
		stHexNumber[idx] = EOS;
		sscanf(stHexNumber, "%lx", pulHexNum);
		*ppc = p;
	}

	return(rc);
}

bool skipLeadingPeriods(char **ppc)
{
	char *p = *ppc;
	int idx = 0;

	if(NULL == p)
		return(false);

	/* skip white space */
	while(isspace(*p) || '.' == *p)
		p++;

	*ppc = p;
	return(true);
}
